import boto3, duckdb, json, os, pandas as pd, pickle, pyarrow as pa, pyarrow.parquet as pq, requests, sys, uuid
from botocore.config import Config
from boto3.s3.transfer import TransferConfig
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional
from pybloom_live import ScalableBloomFilter

from isaura.helpers import (
  MINIO_ENDPOINT,
  MINIO_LOCAL_AK,
  MINIO_LOCAL_SK,
  MAX_ROWS_PER_FILE,
  CHECKPOINT_EVERY,
  BLOOM_FILENAME,
  ACCESS_FILE,
  INDEX_FILE,
  DEFAULT_BUCKET_NAME as pub_bucket,
  DEFAULT_PRIVATE_BUCKET_NAME as priv_bucket,
  logger,
  get_acc_key,
  get_base,
  get_coll,
  get_idx_key,
  get_files_glob,
  get_pref,
  group_inputs,
  hive_prefix,
  make_temp,
  post_apprx,
  query,
  cpu_cnt,
)

class _S3RangeFile:
    def __init__(self, client, bucket, key, size, block_size=262144):
        self._client = client
        self._bucket = bucket
        self._key = key
        self._size = int(size or 0)
        self._block_size = int(block_size)
        self._pos = 0
        self._cache = {}
        self.closed = False

    def readable(self):
        return True

    def seekable(self):
        return True

    def tell(self):
        return self._pos

    def seek(self, offset, whence=0):
        if whence == 0:
            new_pos = int(offset)
        elif whence == 1:
            new_pos = int(self._pos + offset)
        elif whence == 2:
            new_pos = int(self._size + offset)
        else:
            raise ValueError("invalid whence")
        if new_pos < 0:
            new_pos = 0
        if new_pos > self._size:
            new_pos = self._size
        self._pos = new_pos
        return self._pos

    def _get_block(self, block_idx):
        if block_idx in self._cache:
            return self._cache[block_idx]
        start = block_idx * self._block_size
        if start >= self._size:
            data = b""
            self._cache[block_idx] = data
            return data
        end = min(self._size - 1, start + self._block_size - 1)
        resp = self._client.get_object(
            Bucket=self._bucket, Key=self._key, Range=f"bytes={start}-{end}"
        )
        body = resp["Body"].read()
        try:
            resp["Body"].close()
        except Exception:
            pass
        self._cache[block_idx] = body
        return body

    def read(self, n=-1):
        if self.closed:
            return b""
        if n is None or n < 0:
            n = self._size - self._pos
        n = int(n)
        if n <= 0 or self._pos >= self._size:
            return b""
        end_pos = min(self._size, self._pos + n)
        out = bytearray()
        while self._pos < end_pos:
            block_idx = self._pos // self._block_size
            block_off = self._pos % self._block_size
            block = self._get_block(block_idx)
            if not block:
                break
            take = min(len(block) - block_off, end_pos - self._pos)
            if take <= 0:
                break
            out += block[block_off : block_off + take]
            self._pos += take
        return bytes(out)

    def close(self):
        self.closed = True
        self._cache.clear()


class MinioStore:
  def __init__(
    self,
    endpoint=None,
    access=None,
    secret=None,
    multipart_threshold=5 * 1024 * 1024,
    multipart_chunksize=5 * 1024 * 1024,
    max_concurrency=2,
    use_threads=True,
  ):
    self.endpoint = endpoint or MINIO_ENDPOINT
    self.access = access or MINIO_LOCAL_AK
    self.secret = secret or MINIO_LOCAL_SK
    self.client = boto3.client(
      "s3",
      endpoint_url=self.endpoint,
      aws_access_key_id=self.access,
      aws_secret_access_key=self.secret,
      region_name="us-east-1",
      config=Config(signature_version="s3v4", s3={"addressing_style": "path"}),
    )
    self.transfer_config = TransferConfig(
      multipart_threshold=multipart_threshold,
      multipart_chunksize=multipart_chunksize,
      max_concurrency=max_concurrency,
      use_threads=use_threads,
    )
    if not self.ping(self.client):
      sys.exit(1)

  def ping(self, store):
    try:
      url = f"{self.endpoint.rstrip('/')}/minio/health/ready"
      resp = requests.get(url, timeout=3)

      if resp.status_code == 200:
        logger.info(f"MinIO server healthy at {url}")
        return True
      logger.error(
        f"MinIO health check failed. Maybe the minio containers not running[{resp.status_code}]: {resp.text.strip()}"
      )
    except requests.exceptions.RequestException as e:
      logger.error(f"MinIO health request error. Maybe the minio containers not running: {e}")
    except Exception as e:
      logger.error(f"Unexpected error during MinIO health check: {e}")
    return False

  def ensure_bucket(self, bucket):
    try:
      self.client.head_bucket(Bucket=bucket)
    except Exception as e:
      logger.error(e)
      try:
        self.client.create_bucket(Bucket=bucket)
      except Exception as e:
        logger.error(e)
        sys.exit(1)

  def download_file(self, bucket, key, local):
    try:
      self.client.download_file(bucket, key, local, Config=self.transfer_config)
    except Exception as e:
      logger.warning(f"The file {key} is not found in bucket {bucket}. Details -> {e}")
      raise

  def upload_file(self, local, bucket, key, extra_args=None):
    self.client.upload_file(local, bucket, key, ExtraArgs=extra_args or {}, Config=self.transfer_config)

  def list_keys(self, bucket, prefix):
    p = self.client.get_paginator("list_objects_v2")
    for page in p.paginate(Bucket=bucket, Prefix=prefix):
      for obj in page.get("Contents", []):
        yield obj

  def delete_prefix(self, bucket, prefix):
    batch = []
    deleted = 0
    for obj in self.list_keys(bucket, prefix):
      batch.append({"Key": obj["Key"]})
      if len(batch) == 1000:
        self.client.delete_objects(Bucket=bucket, Delete={"Objects": batch})
        deleted += len(batch)
        batch = []
    if batch:
      self.client.delete_objects(Bucket=bucket, Delete={"Objects": batch})
      deleted += len(batch)
    return deleted


class DuckDBMinio:  # Singleton design here
  _instance = None

  def __new__(cls, *a, **kw):
    if cls._instance is None:
      cls._instance = super().__new__(cls)
    return cls._instance

  def __init__(self, endpoint=None, access=None, secret=None, threads=None):
    endpoint = endpoint or MINIO_ENDPOINT
    access = access or MINIO_LOCAL_AK
    secret = secret or MINIO_LOCAL_SK
    threads = threads or max(2, os.cpu_count() or 2)

    cfg = (endpoint, access, secret, threads)
    if getattr(self, "_cfg", None) == cfg:
      return

    if getattr(self, "con", None):
      self.con.close()

    self.endpoint, self.access, self.secret = endpoint, access, secret
    self.con = duckdb.connect(":memory:")
    self.con.execute("PRAGMA enable_object_cache")
    self.con.execute(f"PRAGMA threads={cpu_cnt(ratio=1)};")
    self.con.execute("INSTALL httpfs; LOAD httpfs;")
    self.con.execute("SET s3_access_key_id=?", [access])
    self.con.execute("SET s3_secret_access_key=?", [secret])
    self.con.execute("SET s3_endpoint=?", [endpoint.replace("http://", "").replace("https://", "")])
    self.con.execute("SET s3_region='us-east-1';")
    self.con.execute("SET s3_use_ssl=?", [endpoint.startswith("https://")])
    self.con.execute("SET s3_url_style='path'")
    self._cfg = cfg

  def close(self):
    if getattr(self, "con", None):
      self.con.close()
      self.con = None


class BloomIndex:
  def __init__(
    self,
    store,
    bucket,
    base_prefix,
    local_dir,
    bloom_filename=BLOOM_FILENAME,
    error_rate=1e-14,
    initial_capacity=1_000_000,
  ):
    self.store = store
    self.bucket = bucket
    self.base = base_prefix.strip("/")
    self.bloom_key = f"{self.base}/{bloom_filename}"
    self.index_key = f"{self.base}/{INDEX_FILE}"
    self.local_bloom = os.path.join(local_dir, bloom_filename)
    self.local_index = os.path.join(local_dir, "index.json")
    os.makedirs(local_dir, exist_ok=True)
    try:
      self.store.download_file(self.bucket, self.bloom_key, self.local_bloom)
      with open(self.local_bloom, "rb") as f:
        self.sbf = pickle.load(f)
      logger.info(f"loaded bloom {self.bucket}/{self.bloom_key}")
    except Exception:
      self.sbf = ScalableBloomFilter(
        mode=ScalableBloomFilter.SMALL_SET_GROWTH,
        initial_capacity=initial_capacity,
        error_rate=error_rate,
      )
      logger.info("created new bloom")
    try:
      self.store.download_file(self.bucket, self.index_key, self.local_index)
      with open(self.local_index, "r", encoding="utf-8") as f:
        self.index = json.load(f)
      logger.info(f"loaded index {self.bucket}/{self.index_key} entries={len(self.index)}")
    except Exception:
      self.index = {}
      logger.info("created new index")
    self._added = 0

  def seen(self, v):
    return v in self.sbf

  def seen_many(self, vs):
    sbf = self.sbf
    b = self.bucket
    return {v: [v in sbf, b] for v in vs}

  def rc(self, v):
    return self.index.get(v)

  def register(self, v, rc=None):
    self.sbf.add(v)
    if rc is not None and v not in self.index:
      self.index[v] = list(rc)
    self._added += 1
    if self._added >= CHECKPOINT_EVERY:
      self.persist()

  def persist(self):
    tb = f"{self.local_bloom}.tmp"
    with open(tb, "wb") as f:
      pickle.dump(self.sbf, f, protocol=pickle.HIGHEST_PROTOCOL)
    os.replace(tb, self.local_bloom)
    try:
      self.store.upload_file(self.local_bloom, self.bucket, self.bloom_key)
    except Exception as e:
      logger.warning(f"bloom upload failed: {e}")
    ti = f"{self.local_index}.tmp"
    with open(ti, "w", encoding="utf-8") as f:
      json.dump(self.index, f, separators=(",", ":"), ensure_ascii=False)
    os.replace(ti, self.local_index)
    try:
      self.store.upload_file(self.local_index, self.bucket, self.index_key)
    except Exception as e:
      logger.warning(f"index upload failed: {e}")
    self._added = 0


class TrancheState:
  def __init__(self, store, bucket, base_prefix, tmpdir, max_rows):
    self.store = store
    self.bucket = bucket
    self.base = base_prefix.strip("/")
    self.tmpdir = tmpdir
    self.max_rows = max_rows
    self.state = {}

  def _rows_in_remote(self, key):
    local = os.path.join(self.tmpdir, f"inspect_{uuid.uuid4().hex}.parquet")
    try:
      self.store.download_file(self.bucket, key, local)
      try:
        return pq.ParquetFile(local).metadata.num_rows
      except:
        return len(pd.read_parquet(local))
    except:
      return 0
    finally:
      try:
        os.remove(local)
      except:
        pass

  def ensure(self):
    keys, t = self._list_chunks(), "data"
    if not keys:
      self.state[t] = {"next": 1, "open": None, "rows": 0}
      logger.info(f"tranche new: next=1")
      return

    last = keys[-1]
    n = self._rows_in_remote(last)
    idx = self._chunk_idx(last)

    if n < self.max_rows:
      self.state[t] = {"next": idx, "open": last, "rows": n}
      logger.info(f"tranche open: idx={idx} rows={n}")
    else:
      self.state[t] = {"next": idx + 1, "open": None, "rows": 0}
      logger.info(f"tranche rotate: next={idx + 1}")

  def _list_chunks(self):
    pref = hive_prefix(self.base) + "/"
    keys = []
    for obj in self.store.list_keys(self.bucket, pref):
      k = obj["Key"]
      if k.endswith(".parquet") and "/chunk_" in k:
        keys.append(k)
    return sorted(keys, key=self._chunk_idx)

  def _chunk_idx(self, key):
    base = os.path.basename(key)
    name, _ = os.path.splitext(base)
    try:
      return int(name.split("_")[1])
    except:
      return 1

  def _write_chunk(self, df, idx, mode="new", existing_local=None):
    os_key = f"{hive_prefix(self.base)}/chunk_{idx}.parquet"
    local = existing_local or os.path.join(self.tmpdir, f"chunk_{uuid.uuid4().hex}.parquet")

    if mode == "append" and existing_local:
      old = pd.read_parquet(existing_local)
      df = pd.concat([old, df], ignore_index=True)

    df.to_parquet(local, index=False)
    self.store.upload_file(local, self.bucket, os_key)

    if not existing_local:
      try:
        os.remove(local)
      except:
        pass

    return os_key

  def flush(self, rows, schema_cols):
    if not rows:
      return self.ensure()
    self.ensure()
    st = self.state["data"]
    df_all = pd.DataFrame(rows)

    for col in schema_cols:
      if col not in df_all.columns:
        df_all[col] = pd.Series([None] * len(df_all))
    df_all = df_all[schema_cols]

    remaining = len(df_all)
    start = 0

    logger.info(f"flush: tranche rows={remaining}")

    if st["open"]:
      tmp = os.path.join(self.tmpdir, f"open_{uuid.uuid4().hex}.parquet")
      try:
        self.store.download_file(self.bucket, st["open"], tmp)
        space = self.max_rows - st["rows"]
        take = min(space, remaining)

        if take > 0:
          part = df_all.iloc[start : start + take]
          self._write_chunk(part, st["next"], mode="append", existing_local=tmp)
          st["rows"] += take
          remaining -= take
          start += take
          logger.info(f"flush: appended tranche idx={st['next']} +{take} -> {st['rows']}")

        if st["rows"] >= self.max_rows:
          st["next"] += 1
          st["open"] = None
          st["rows"] = 0
          logger.info(f"flush: closed tranche next={st['next']}")
      finally:
        try:
          os.remove(tmp)
        except:
          pass

    while remaining > 0:
      take = min(self.max_rows, remaining)
      part = df_all.iloc[start : start + take]
      os_key = self._write_chunk(part, st["next"], mode="new")

      if take < self.max_rows:
        st["open"] = os_key
        st["rows"] = take
        logger.info(f"flush: new open tranche idx={st['next']} rows={take}")
      else:
        st["next"] += 1
        st["open"] = None
        st["rows"] = 0
        logger.info(f"flush: full tranche idx={st['next'] - 1} rows={take}")

      remaining -= take
      start += take


@dataclass
class AddRowsStats:
  in_count: int
  added: int = 0
  skipped_seen: int = 0
  skipped_blank: int = 0
  skipped_error: int = 0
  sample_seen: List[Any] = field(default_factory=list)
  sample_errors: List[Dict[str, Any]] = field(default_factory=list)
  new_rows: List[Dict[str, Any]] = field(default_factory=list)


class _SinkWriter:
  def __init__(self, store, bucket, model_id, model_version, tmpdir):
    self.store = store
    self.bucket = bucket
    self.model_id = model_id
    self.model_version = model_version
    self.base = get_base(model_id, model_version)
    self.tmpdir = tmpdir
    self.max_rows = MAX_ROWS_PER_FILE
    self.store.ensure_bucket(self.bucket)
    self.bi = BloomIndex(self.store, self.bucket, self.base, tmpdir)
    self.tranche = TrancheState(self.store, self.bucket, self.base, tmpdir, self.max_rows)
    self.buffers = defaultdict(list)

  def add_rows(self, r, c, df):
    try:
      if df is None or df.empty:
        logger.info(f"[add_rows] tranche=({r},{c}) df is empty -> added=0")
        return 0

      stats = AddRowsStats(in_count=len(df))

      logger.info(
        f"[add_rows] tranche=({r},{c}) in={stats.in_count} "
        f"bucket={getattr(self, 'bucket', None)} bloom_bucket={getattr(self.bi, 'bucket', None)}"
      )

      for i, row in df.iterrows():
        smi: Optional[str] = None
        try:
          smi = row.get("input") if hasattr(row, "get") else row["input"]
          smi = (smi or "").strip() if isinstance(smi, str) else smi

          if not smi:
            stats.skipped_blank += 1
            continue

          if not self.bi.seen(smi):
            stats.new_rows.append(dict(row))
            self.bi.register(smi, rc=(r, c))
            stats.added += 1
          else:
            stats.skipped_seen += 1
            if len(stats.sample_seen) < 3:
              stats.sample_seen.append(smi)

        except Exception as e:
          stats.skipped_error += 1
          if len(stats.sample_errors) < 3:
            stats.sample_errors.append({"row_idx": i, "smi": smi, "err": repr(e)})
          logger.info(f"[add_rows] tranche=({r},{c}) row_idx={i} smi={smi!r} -> skipped due to error: {e}")

      if not stats.new_rows:
        logger.info(
          f"[add_rows] tranche=({r},{c}) in={stats.in_count} added=0 "
          f"skipped_seen={stats.skipped_seen} skipped_blank={stats.skipped_blank} skipped_error={stats.skipped_error} "
          f"{'sample_seen=' + str(stats.sample_seen) if stats.sample_seen else ''} "
          f"{'sample_errors=' + str(stats.sample_errors) if stats.sample_errors else ''}"
        )
        return 0

      try:
        self.buffers[(r, c)].extend(stats.new_rows)
      except Exception as e:
        logger.info(
          f"[add_rows] tranche=({r},{c}) failed to extend buffer with {len(stats.new_rows)} rows: {e}"
        )
        return stats.added

      buf_len = len(self.buffers[(r, c)])
      logger.info(
        f"[add_rows] tranche=({r},{c}) in={stats.in_count} added={stats.added} "
        f"skipped_seen={stats.skipped_seen} skipped_blank={stats.skipped_blank} skipped_error={stats.skipped_error} "
        f"buffer={buf_len}/{self.max_rows}"
      )

      if buf_len >= self.max_rows:
        try:
          logger.info(f"[add_rows] tranche=({r},{c}) buffer reached max_rows -> flushing {buf_len} rows")
          self.tranche.flush(self.buffers[(r, c)], list(df.keys()))
          self.buffers[(r, c)].clear()
          logger.info(f"[add_rows] tranche=({r},{c}) flush complete -> buffer cleared")
        except Exception as e:
          logger.error(f"[add_rows] tranche=({r},{c}) flush FAILED (buffer kept, len={buf_len}): {e}")

      return stats.added

    except Exception as e:
      logger.error(f"[add_rows] tranche=({r},{c}) FATAL error: {e}")
      return 0

  def finalize(self, metadata_local=None, schema_cols=None):
    for k in list(self.buffers.keys()):
      if self.buffers[k]:
        self.tranche.flush(self.buffers[k], schema_cols)
        self.buffers[k].clear()
    self.bi.persist()
    if metadata_local:
      try:
        self.store.upload_file(metadata_local, self.bucket, f"{self.base}/{ACCESS_FILE}")
      except Exception:
        pass


class _BaseTransfer:
  def __init__(self, model_id, model_version, bucket, output_dir=None):
    self.model_id = model_id
    self.model_version = model_version
    self.bucket = bucket
    self.output_dir = output_dir
    self.base = get_pref(model_id, model_version)
    self.tranches = get_base(model_id, model_version)
    self.collection = get_coll(self.model_id, self.model_version)
    self.store = MinioStore()
    self.tmpdir = make_temp("isaura_xfer_")
    self.tmpdir_sinkw = make_temp("isaura_sinkw_")
    self.duck = DuckDBMinio(
      endpoint=self.store.endpoint,
      access=self.store.access,
      secret=self.store.secret,
    )
    self.bi = BloomIndex(
      self.store,
      self.bucket,
      self.tranches,
      self.tmpdir,
      bloom_filename=os.getenv("BLOOM_FILENAME", BLOOM_FILENAME),
    )

  def _download_if_exists(self, key, local):
    try:
      self.store.download_file(self.bucket, key, local)
      return True
    except Exception as e:
      logger.error(f"Either model id or version or project name are maybe specified wrongly. Results -> {e}")
      sys.exit(1)
      return False

  def _load_metadata(self):
    local = os.path.join(self.tmpdir, ACCESS_FILE)
    acc = get_acc_key(self.tranches)
    try:
      if self._download_if_exists(acc, local):
        with open(local, "r", encoding="utf-8") as f:
          return local, json.load(f)
    except Exception as e:
      logger.error(f"{e}. Possible causes -> 1. No project 2. Incorrect model id or version")

  def _load_index(self):
    key = get_idx_key(self.tranches)
    local = os.path.join(self.tmpdir, "index_src.json")
    if not self._download_if_exists(key, local):
      raise RuntimeError("index.json not found")
    with open(local, "r", encoding="utf-8") as f:
      return json.load(f), local

  def select_rows(self, wanted, input_col="input"):
    return query(self.duck.con, input_col, wanted, get_files_glob(self.bucket, self.tranches))

  def _delete(self):
    return self.store.delete_prefix(self.bucket, self.tranches)

  def _pull(self, df, index):
    input = df["input"].astype(str).to_list()
    gp = group_inputs(input, index, bloom=self.bi, force=True)
    wt = _SinkWriter(self.store, self.bucket, self.model_id, self.model_version, self.tmpdir)
    tp, tu, dfs = 0, 0, []
    if gp is None:
      return 0, 0
    for (r, c), want in gp.items():
      df_ = df[df["input"].isin(list(want))]
      _tp = wt.add_rows(r, c, df)
      if _tp != 0:
        dfs.append(df_)
      tp += _tp
    if wt:
      wt.finalize(schema_cols=list(df.keys()))
    if dfs:
      dfs = pd.concat(dfs, ignore_index=True)
      post_apprx(dfs, self.collection)
    return tp, tu

  def _dump(self):
    os.makedirs(self.output_dir, exist_ok=True)
    paginator = self.store.client.get_paginator("list_objects_v2")
    pages = paginator.paginate(Bucket=self.bucket)
    found = False
    for page in pages:
      for obj in page.get("Contents", []):
        found = True
        key = obj["Key"]
        local_path = os.path.join(self.output_dir, key)
        os.makedirs(os.path.dirname(local_path), exist_ok=True)
        logger.info(f"dumping from minio://{self.bucket}/{key} -> {local_path}")
        self.store.download_file(self.bucket, key, local_path)
    if not found:
      logger.info(f"bucket empty: {self.bucket}")

  def _copy(self, meta_local, meta_list):
    if self.output_dir is not None:
      self._dump()
      return
    priv = [
      (d.get("input") or "").strip()
      for d in meta_list
      if (d.get("access") or "").lower() == "private" and (d.get("input") or "").strip()
    ]
    pub = [
      (d.get("input") or "").strip()
      for d in meta_list
      if (d.get("access") or "").lower() == "public" and (d.get("input") or "").strip()
    ]
    index, _ = self._load_index()
    self.bi.bucket = priv_bucket
    gp = group_inputs(priv, index, bloom=self.bi) if priv else {}
    self.bi.bucket = pub_bucket
    gu = group_inputs(pub, index, bloom=self.bi) if pub else {}
    w_priv = (
      _SinkWriter(self.store, priv_bucket, self.model_id, self.model_version, self.tmpdir_sinkw)
      if gp
      else None
    )
    w_pub = (
      _SinkWriter(self.store, pub_bucket, self.model_id, self.model_version, self.tmpdir_sinkw)
      if gu
      else None
    )
    tp, tu, dfs = 0, 0, []
    for (r, c), want in gp.items():
      df = self.select_rows(want)
      _tp = w_priv.add_rows(r, c, df)
      if _tp != 0:
        dfs.append(df)
      tp += _tp
    for (r, c), want in gu.items():
      df = self.select_rows(want)
      _tu = w_pub.add_rows(r, c, df)
      if _tu != 0:
        dfs.append(df)
      tu += _tu
    if w_priv:
      w_priv.finalize(metadata_local=meta_local, schema_cols=list(df.keys()))
    if w_pub:
      w_pub.finalize(metadata_local=meta_local, schema_cols=list(df.keys()))
    if dfs:
      dfs = pd.concat(dfs, ignore_index=True)
      post_apprx(dfs, self.collection)
    return tp, tu



import csv, datetime, json, os, shutil, sys, time, uuid, pyarrow.parquet as pq, pandas as pd
from collections import defaultdict, Counter
from collections import defaultdict
from contextlib import AbstractContextManager
from isaura.base import _BaseTransfer, BloomIndex, TrancheState, MinioStore, DuckDBMinio, _S3RangeFile
from isaura.helpers import (
  ACCESS_FILE,
  BLOOM_FILENAME,
  CHECKPOINT_EVERY,
  DEFAULT_BUCKET_NAME as PUB,
  DEFAULT_PRIVATE_BUCKET_NAME as PRI,
  INPUT_C,
  MAX_ROWS,
  MINIO_ENDPOINT_CLOUD,
  MIN_NNS_RESULT_SIZE,
  MINIO_CLOUD_AK as mcak,
  MINIO_CLOUD_SK as mcsk,
  MINIO_PRIV_CLOUD_AK as mcpak,
  MINIO_PRIV_CLOUD_SK as mcpsk,
  logger,
  fetch_schema_from_github,
  get_acc_key,
  get_apprx,
  get_base,
  get_coll,
  get_files_glob,
  get_idx_key,
  get_pref,
  group_inputs,
  make_temp,
  query,
  split_csv,
  spinner,
  tranche_coordinates,
  track_write_progress,
  write_access_file,
)


class IsauraChecker(AbstractContextManager):
  def __init__(
    self,
    bucket,
    model_id,
    model_version,
    store=None,
    bloom_filename=BLOOM_FILENAME,
    checkpoint_every=CHECKPOINT_EVERY,
  ):
    self.bucket = bucket
    self.base_prefix = get_base(model_id, model_version)
    self.store = store or MinioStore()
    self.local_dir = make_temp("isaura_checker_")
    self.checkpoint_every = checkpoint_every
    self.bi = BloomIndex(
      self.store,
      self.bucket,
      self.base_prefix,
      self.local_dir,
      bloom_filename=bloom_filename,
    )
    self._added = 0

  def seen(self, v):
    return self.bi.seen(v)

  def seen_many(self, vs):
    return self.bi.seen_many(vs)

  def register(self, v, rc=None):
    self.bi.register(v, rc=rc)
    self._added += 1
    if self._added >= self.checkpoint_every:
      self.bi.persist()
      self._added = 0

  def close(self):
    if self._added:
      self.bi.persist()
      self._added = 0

  def __enter__(self):
    return self

  def __exit__(self, et, ev, tb):
    self.close()


class IsauraWriter:
  def __init__(
    self,
    input_csv,
    model_id,
    model_version,
    access="public",
    bucket=None,
    endpoint=None,
    access_key=None,
    secrete=None,
  ):
    self.input_csv = input_csv
    self.model_id = model_id
    self.model_version = model_version
    self.access = access
    self.bucket = bucket or PUB
    self.base_prefix = get_base(self.model_id, model_version)
    self.access_key = get_acc_key(self.base_prefix)
    self.max_rows = int(MAX_ROWS)
    self.store = MinioStore(endpoint=endpoint, access=access_key, secret=secrete)
    self.store.ensure_bucket(self.bucket)
    self.tmpdir = make_temp("isaura_writter_")
    self.metadata_path = os.path.join(self.tmpdir, ACCESS_FILE)
    self.bi = BloomIndex(
      self.store,
      self.bucket,
      self.base_prefix,
      self.tmpdir,
      bloom_filename=os.getenv("BLOOM_FILENAME", BLOOM_FILENAME),
    )
    self.tranche = TrancheState(self.store, self.bucket, self.base_prefix, self.tmpdir, self.max_rows)
    self.buffers = defaultdict(list)
    self.schema_cols = None
    logger.info(f"writer init: bucket={self.bucket} base={self.base_prefix} csv={self.input_csv}")

  def _load_metadata(self):
    local = os.path.join(self.tmpdir, f"{uuid.uuid4().hex}.json")
    try:
      self.store.download_file(self.bucket, self.access_key, local)
      with open(local, "r", encoding="utf-8") as f:
        return json.load(f)
    except Exception:
      return None
    finally:
      try:
        os.remove(local)
      except Exception:
        pass

  def _upload_metadata(self, inputs):
    if not self.metadata_path or not self.bucket:
      return
    try:
      existed = self._load_metadata()
      write_access_file(existed, inputs, self.access, self.metadata_path)
      self.store.upload_file(self.metadata_path, self.bucket, f"{self.base_prefix}/{ACCESS_FILE}")
      logger.info(f"{ACCESS_FILE} -> minio://{self.bucket}/{self.base_prefix}/{ACCESS_FILE}")
    except Exception as e:
      logger.warning(f"metadata upload failed: {e}")

  def _set_schema(self, row):
    if self.schema_cols is None:
      self.schema_cols = list(row.keys())
      logger.info(f"writer schema: {self.schema_cols[:10]}")

  def _flush_if_needed(self, r, c):
    buf = self.buffers[(r, c)]
    if len(buf) >= self.max_rows:
      self.tranche.flush(buf, self.schema_cols)
      self.buffers[(r, c)].clear()

  def write(self, df=None, show_progress: bool = True):
    total = dupes = 0
    new, buffs = [], []

    rows_iter, total_rows = None, None

    if df is not None:
      rows_list = df.to_dict("records")
      rows_iter = rows_list
      total_rows = len(rows_list)
    else:
      f = open(self.input_csv, newline="", encoding="utf-8")
      reader = csv.DictReader(f)
      rows_iter = reader
      total_rows = None

    try:
      if show_progress:
        rows_iter = track_write_progress(
          rows_iter,
          total=total_rows,
          description="Writing rows",
          console=logger.console,
        )

      for row in rows_iter:
        self._set_schema(row)
        smi = (row.get("input") or row.get("smiles") or "").strip()
        if not smi:
          continue

        if not self.bi.seen(smi):
          try:
            r, c, _, _ = tranche_coordinates(smi)
            if r < 1 or c < 1:
              raise ValueError("Tranche coordinates must be >= 1")
          except Exception:
            logger.warning("invalid SMILES or tranche mapping; skipped")
            continue

          new.append(smi)
          self.buffers[(r, c)].append(dict(row))
          self.bi.register(smi, rc=(r, c))
          total += 1

          self._flush_if_needed(r, c)
          if self.bi._added >= int(CHECKPOINT_EVERY):
            self.bi.persist()
        else:
          dupes += 1
          continue

    finally:
      if df is None and "f" in locals():
        try:
          f.close()
        except Exception:
          pass

    for (r, c), buf in list(self.buffers.items()):
      if buf:
        buffs.extend(buf)
        self.buffers[(r, c)].clear()

    self.tranche.flush(buffs, self.schema_cols)
    self.bi.persist()

    if new:
      self._upload_metadata(new)

    logger.info(f"write done: new={total} dupes={dupes}")

  def close(self):
    try:
      shutil.rmtree(self.tmpdir)
    except Exception:
      pass

  def __enter__(self):
    return self

  def __exit__(self, et, ev, tb):
    self.close()


class IsauraReader:
  def __init__(
    self,
    model_id,
    model_version,
    input_csv,
    approximate,
    bucket=None,
    endpoint=None,
    access_key=None,
    secrete=None,
  ):
    self.model_id = model_id
    self.model_version = model_version
    self.approximate = approximate
    self.input_csv = input_csv
    self.bucket = bucket or PUB
    self.endpoint = endpoint
    self.base = get_base(self.model_id, model_version)
    self.pref = get_pref(self.model_id, self.model_version)
    self.collection = get_coll(self.model_id, self.model_version)
    self.index_key = get_idx_key(self.base)
    self.tmpdir = make_temp("isaura_reader_")
    self.store = MinioStore(endpoint=self.endpoint, access=access_key, secret=secrete)
    self.duck = DuckDBMinio(endpoint=self.endpoint, access=access_key, secret=secrete)
    self.bi = BloomIndex(
      self.store,
      self.bucket,
      self.base,
      self.tmpdir,
      bloom_filename=os.getenv("BLOOM_FILENAME", BLOOM_FILENAME),
    )
    logger.info(f"reader init bucket={self.bucket} base={self.base} csv={self.input_csv}")

  def _load_index(self):
    local = os.path.join(self.tmpdir, f"{uuid.uuid4().hex}.json")
    self.store.download_file(self.bucket, self.index_key, local)
    try:
      with open(local, "r", encoding="utf-8") as f:
        return json.load(f)
    except Exception as e:
      logger.error(e)
    finally:
      try:
        os.remove(local)
      except:
        pass

  def read(self, output_csv=None, df=None):
    index = self._load_index()
    t0, wanted, header_set = time.time(), [], set()
    rows = df.to_dict("records") if df is not None else None
    if rows is None:
      with open(self.input_csv, newline="", encoding="utf-8") as f:
        rows = list(csv.DictReader(f))
    for row in rows:
      h = INPUT_C[0] if row.get(INPUT_C[0]) else INPUT_C[1]
      v = (row.get(h) or "").strip()
      if v:
        wanted.append(v)
      if h:
        header_set.add(h)
    if index and len(index) < MIN_NNS_RESULT_SIZE and self.approximate:
      logger.error(
        f"Minimum precalculation size for enabling nearest neighbor search is {MIN_NNS_RESULT_SIZE}, found {len(index)}. Aborting the Ops!"
      )
      sys.exit(1)
    if self.approximate:
      st = time.perf_counter()
      wanted = get_apprx(wanted, self.collection)
      et = time.perf_counter()
      logger.info(f"Approximate inputs are retrieved {len(wanted)} in {et - st:.2f} seconds!")
    header = list(header_set)[0] if header_set else "smiles"
    group_inputs(wanted, index, bloom=self.bi)
    if not wanted:
      return pd.DataFrame()
    try:
      files = get_files_glob(self.bucket, self.base)
      out = spinner("Fetching queries. Please wait!", query, self.duck.con, header, wanted, files)
      elapsed = time.time() - t0
      logger.success(f"Query successfully fetched for a given inputs in {elapsed:.2f} sec")
    except Exception as e:
      logger.error(e)
      sys.exit(1)
    rate = (len(out) / elapsed) if elapsed > 0 and len(out) else 0.0
    if output_csv:
      logger.debug(f"writting csv rows={len(out)} path={output_csv}. Please wait!")
      out.to_csv(output_csv, index=False)
      logger.info(f"wrote csv rows={len(out)} path={output_csv}")
    te = time.time() - t0
    logger.info(
      f"read done model+version={self.pref} "
      f"bucket={self.bucket} inputs={len(wanted)} matched={len(out)} "
      f"(query elapsed)={elapsed:.2f}s rate={rate:.1f}/s (total time elapsed)={te:.2f}s"
    )
    return out


class IsauraCopy(_BaseTransfer):
  def copy(self):
    meta_local, meta = self._load_metadata()
    return self._copy(meta_local, meta)


class IsauraMover(_BaseTransfer):
  def move(self):
    meta_local, meta = self._load_metadata()
    self._copy(meta_local, meta)
    n = self._delete()
    logger.info(f"move wiped objects={n}")


class IsauraRemover(_BaseTransfer):
  def remove(self):
    n = self._delete()
    logger.info(f"removed objects={n}")


class IsauraPull(_BaseTransfer):
  def __init__(self, model_id, model_version, bucket, input_csv, output_dir=None):
    super().__init__(model_id, model_version, bucket, output_dir)
    self.model_id = model_id
    self.model_version = model_version
    self.bucket = bucket
    self.input_csv = input_csv

  def pull(self):
    logger.info(
      f"Pulling precalculation from the cloud for model={self.model_id}, version={self.model_version}, bucket={self.bucket}"
    )
    r = IsauraReader(
      model_id=self.model_id,
      model_version=self.model_version,
      input_csv=self.input_csv,
      approximate=False,
      bucket=self.bucket,
      endpoint=MINIO_ENDPOINT_CLOUD,
      access_key=mcak,
      secrete=mcsk,
    )
    out = self._pull(r.read(), self._load_index())
    logger.info(f"pulled objects={out}")


class IsauraPush:
  def __init__(self, model_id, model_version, bucket):
    self.model_id = model_id
    self.model_version = model_version
    self.bucket = bucket

  def push(self):
    insp = IsauraInspect(
      model_id=self.model_id,
      model_version=self.model_version,
      access="both",
      cloud=False,
    )
    df = insp.list_available()
    if df.empty:
      logger.error("No data found in any default bucket for a given model! Aborting push.")
      sys.exit(1)

    files = split_csv(df)

    if not files:
      logger.error("No data found in any default bucket! Aborting push.")
      sys.exit(1)

    file1 = files[0]
    file2 = files[1] if len(files) > 1 else None

    if not file2:
      logger.warning("Private bucket has no data! Skipping push for it.")

    for access, file, mck, mcs in [("public", file1, mcak, mcsk), ("private", file2, mcpak, mcpsk)]:
      if not file:
        continue
      r = IsauraReader(
        model_id=self.model_id,
        model_version=self.model_version,
        input_csv=file,
        approximate=False,
        bucket=f"isaura-{access}",
      )
      df = r.read()
      with IsauraWriter(
        input_csv=file,
        model_id=self.model_id,
        model_version=self.model_version,
        bucket=f"isaura-{access}",
        access=None if access == "public" else "private",
        endpoint=MINIO_ENDPOINT_CLOUD,
        access_key=mck,
        secrete=mcs,
      ) as w:
        w.write(df=df)



class IsauraInspect:
    def __init__(
        self,
        model_id=None,
        model_version=None,
        cloud=False,
        project_name=None,
        access="both",
        endpoint=None,
        parquet_block_size=262144,
        max_small_json_bytes=20000000,
        heavy_index=False,
    ):
        self.model_id, self.model_version = model_id, model_version
        self.cloud, self.project_name, self.access = cloud, project_name, access
        self.endpoint = MINIO_ENDPOINT_CLOUD if cloud else endpoint
        self.parquet_block_size = int(parquet_block_size)
        self.max_small_json_bytes = int(max_small_json_bytes)
        self.heavy_index = bool(heavy_index)
        self._cache = {}
        self._rowcount_cache = {}

    def buckets(self):
        if self.project_name:
            return [self.project_name]
        if self.access == "public":
            return [PUB]
        if self.access == "private":
            return [PRI]
        return [PUB, PRI]

    def _creds(self, bucket):
        if not self.cloud:
            return {"endpoint": self.endpoint}
        if bucket == PUB:
            return {"endpoint": MINIO_ENDPOINT_CLOUD, "access": mcak, "secret": mcsk}
        if bucket == PRI:
            return {"endpoint": MINIO_ENDPOINT_CLOUD, "access": mcpak, "secret": mcpsk}
        return {"endpoint": MINIO_ENDPOINT_CLOUD}

    def _clients(self, bucket):
        k = (bucket, self.cloud, self.endpoint)
        if k not in self._cache:
            c = self._creds(bucket)
            self._cache[k] = (MinioStore(**c),)
        return self._cache[k]

    def _client(self, bucket):
        return self._clients(bucket)[0].client

    def _paginator(self, bucket):
        return self._client(bucket).get_paginator("list_objects_v2")

    def list_prefixes(self, bucket, prefix=""):
        p = self._paginator(bucket)
        for page in p.paginate(Bucket=bucket, Prefix=prefix, Delimiter="/"):
            for cp in page.get("CommonPrefixes", []):
                yield cp["Prefix"]

    def list_objects(self, bucket, prefix=""):
        p = self._paginator(bucket)
        for page in p.paginate(Bucket=bucket, Prefix=prefix):
            for obj in page.get("Contents", []):
                yield obj["Key"]

    def iter_object_meta(self, bucket, prefix=""):
        p = self._paginator(bucket)
        for page in p.paginate(Bucket=bucket, Prefix=prefix):
            for obj in page.get("Contents", []):
                yield {
                    "Key": obj.get("Key", ""),
                    "Size": obj.get("Size", 0),
                    "LastModified": obj.get("LastModified"),
                    "ETag": obj.get("ETag"),
                    "StorageClass": obj.get("StorageClass"),
                }

    def get_json(self, bucket, key, max_bytes=None):
        client = self._client(bucket)
        lim = self.max_small_json_bytes if max_bytes is None else int(max_bytes)
        try:
            try:
                h = client.head_object(Bucket=bucket, Key=key)
                if h and int(h.get("ContentLength", 0) or 0) > lim:
                    return None
            except Exception:
                pass
            o = client.get_object(Bucket=bucket, Key=key)
            body = o["Body"].read()
            try:
                o["Body"].close()
            except Exception:
                pass
            return json.loads(body.decode("utf-8"))
        except Exception:
            return None

    def iter_models(self, bucket, prefix_filter=""):
        if self.model_id:
            m = self.model_id
            if self.model_version:
                yield m, self.model_version
                return
            for v_pref in self.list_prefixes(bucket, f"{m}/"):
                mv = v_pref[len(f"{m}/") :].strip("/")
                if mv:
                    yield m, mv
            return

        for m_pref in self.list_prefixes(bucket, ""):
            mid = m_pref.strip("/")
            if prefix_filter and not mid.startswith(prefix_filter):
                continue
            for v_pref in self.list_prefixes(bucket, m_pref):
                mv = v_pref[len(m_pref) :].strip("/")
                if mv:
                    yield mid, mv

    def load_index(self, bucket, model_id, model_version):
        base = get_base(model_id, model_version)
        key = get_idx_key(base)
        if self.heavy_index or not self.cloud:
            return self.get_json(bucket, key, max_bytes=10**18) or {}
        return {}

    def load_metadata(self, bucket, model_id, model_version):
        base = get_base(model_id, model_version)
        return self.get_json(bucket, f"{base}/{ACCESS_FILE}")

    def _indices_union(self):
        union, owner = {}, {}
        for b in self.buckets():
            for mid, mv in self.iter_models(b):
                idx = self.load_index(b, mid, mv)
                for smi in idx.keys():
                    if smi not in union:
                        union[smi] = True
                        owner[smi] = b
        return union, owner

    def list_available(self, output_csv=None):
        _, owner = self._indices_union()
        df = pd.DataFrame([{"input": smi, "bucket": b} for smi, b in owner.items()])
        if output_csv:
            df.to_csv(output_csv, index=False)
        return df

    def inspect_inputs(self, input_csv, output_csv=None):
        _, owner = self._indices_union()
        with open(input_csv, newline="", encoding="utf-8") as f:
            wanted = [(r.get("input") or "").strip() for r in csv.DictReader(f) if (r.get("input") or "").strip()]
        df = pd.DataFrame([{"input": s, "available": s in owner, "bucket": owner.get(s, "")} for s in wanted])
        if output_csv:
            df.to_csv(output_csv, index=False)
        return df

    def find_chunk_keys(self, bucket, model_id, model_version):
        pref = get_pref(model_id, model_version)
        for obj in self.iter_object_meta(bucket, pref):
            k = obj.get("Key") or ""
            if "/chunk_" in k and k.endswith(".parquet"):
                yield k, int(obj.get("Size") or 0)

    def parquet_num_rows(self, bucket, parquet_key, size=None):
        if pq is None or not parquet_key:
            return None
        ck = (bucket, parquet_key)
        if ck in self._rowcount_cache:
            return self._rowcount_cache[ck]
        client = self._client(bucket)
        try:
            if size is None:
                h = client.head_object(Bucket=bucket, Key=parquet_key)
                size = int(h.get("ContentLength", 0) or 0)
            size = int(size or 0)
            if size <= 0:
                return None
            f = _S3RangeFile(
                client,
                bucket,
                parquet_key,
                size=size,
                block_size=self.parquet_block_size,
            )
            pf = pq.ParquetFile(f)
            md = pf.metadata
            n = int(md.num_rows) if md is not None else None
            try:
                f.close()
            except Exception:
                pass
            if n is not None:
                self._rowcount_cache[ck] = n
            return n
        except Exception:
            return None

    def entries_from_chunks(self, bucket, model_id, model_version):
        total = 0
        n_chunks = 0
        for k, sz in self.find_chunk_keys(bucket, model_id, model_version):
            n_chunks += 1
            n = self.parquet_num_rows(bucket, k, size=sz)
            if n is not None:
                total += int(n)
        return total, n_chunks

    def inspect_models(self, bucket, prefix_filter=""):
        out = []
        for mid, mv in self.iter_models(bucket, prefix_filter=prefix_filter):
            if self.cloud and not self.heavy_index:
                entries, chunks = self.entries_from_chunks(bucket, mid, mv)
                out.append(
                    {
                        "bucket": bucket,
                        "model_id": mid,
                        "model_version": mv,
                        "model": f"{mid}/{mv}",
                        "entries": int(entries),
                        "chunks": int(chunks),
                    }
                )
            else:
                out.append(
                    {
                        "bucket": bucket,
                        "model_id": mid,
                        "model_version": mv,
                        "model": f"{mid}/{mv}",
                        "entries": len(self.load_index(bucket, mid, mv)),
                        "chunks": None,
                    }
                )
        return out

    def find_any_chunk_key(self, bucket, model_id, model_version):
        pref = get_pref(model_id, model_version)
        for k in self.list_objects(bucket, pref):
            if "/chunk_" in k and k.endswith(".parquet"):
                return k
        return None

    def parquet_columns(self, bucket, parquet_key):
        if not parquet_key or pq is None:
            return None
        client = self._client(bucket)
        try:
            h = client.head_object(Bucket=bucket, Key=parquet_key)
            size = int(h.get("ContentLength", 0) or 0)
            if size <= 0:
                return None
            f = _S3RangeFile(
                client,
                bucket,
                parquet_key,
                size=size,
                block_size=self.parquet_block_size,
            )
            pf = pq.ParquetFile(f)
            cols = list(pf.schema_arrow.names)
            try:
                f.close()
            except Exception:
                pass
            return cols
        except Exception:
            return None
        
    def duckdb_columns(self, bucket, parquet_key):
        return self.parquet_columns(bucket, parquet_key)
    

class IsauraStat:
  def __init__(
    self,
    project_name=None,
    access="both",
    cloud=False,
    endpoint=None,
    include_columns=True,
    include_column_names=False,
    schema_version="1",
    producer="isaura stats",
  ):
    self.insp = IsauraInspect(cloud=cloud, project_name=project_name, access=access, endpoint=endpoint)
    self.include_columns = bool(include_columns)
    self.include_column_names = bool(include_column_names)
    self.schema_version = str(schema_version)
    self.producer = str(producer)

  def _percentile(self, sorted_vals, p):
    if not sorted_vals:
      return None
    n = len(sorted_vals)
    i = int(round((p / 100.0) * (n - 1)))
    i = max(0, min(i, n - 1))
    return sorted_vals[i]

  def compute(self):
    generated_at = datetime.datetime.now(datetime.timezone.utc).isoformat()
    buckets = self.insp.buckets()

    models = []
    models_per_molecule = Counter()
    models_total_by_bucket = defaultdict(int)

    for b in buckets:
      for mid, mv in self.insp.iter_models(b):
        idx = self.insp.load_index(b, mid, mv)
        for smi in idx.keys():
          models_per_molecule[smi] += 1

        meta_out = fetch_schema_from_github(mid)

        cols = None
        ncols = None
        if self.include_columns:
          ck = self.insp.find_any_chunk_key(b, mid, mv)
          cols = self.insp.duckdb_columns(b, ck) if ck else None
          ncols = len(cols) if cols else None

        row = {
          "bucket": b,
          "model_id": mid,
          "model_version": mv,
          "model_key": f"{b}:{mid}:{mv}",
          "model": f"{mid}/{mv}",
          "molecules": len(idx),
          "n_columns": ncols,
          "metadata": meta_out,
        }
        if self.include_column_names:
          row["columns"] = cols

        models.append(row)
        models_total_by_bucket[b] += 1

    counts = sorted(models_per_molecule.values())
    hist = Counter(counts)
    hist_out = [{"models": k, "molecules": v} for k, v in sorted(hist.items(), key=lambda x: x[0])]

    return {
      "schema_version": self.schema_version,
      "producer": self.producer,
      "generated_at_utc": generated_at,
      "buckets": buckets,
      "models_total": len(models),
      "models_total_by_bucket": dict(models_total_by_bucket),
      "models": models,
      "models_per_molecule": {
        "molecules_total_unique": len(models_per_molecule),
        "min": min(counts) if counts else None,
        "max": max(counts) if counts else None,
        "p50": self._percentile(counts, 50),
        "p95": self._percentile(counts, 95),
        "histogram": hist_out,
      },
    }

  def write_json(self, output_path):
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    tmp = output_path + ".tmp"
    data = self.compute()
    with open(tmp, "w", encoding="utf-8") as f:
      json.dump(data, f, indent=2, ensure_ascii=False)
    os.replace(tmp, output_path)
    return output_path



import datetime, os, sys

import rich_click as click
import rich_click.rich_click as rc

from isaura.manage import (
  IsauraMover,
  IsauraCopy,
  IsauraRemover,
  IsauraWriter,
  IsauraReader,
  IsauraInspect,
  IsauraPull,
  IsauraPush,
  IsauraStat,
)
from isaura.helpers import (
  DEFAULT_BUCKET_NAME,
  DEFAULT_PRIVATE_BUCKET_NAME,
  logger,
  console,
  inspect_table,
  make_table,
  show_figlet,
  run_docker_compose,
  spinner,
)

click.rich_click.USE_RICH_MARKUP = True
click.rich_click.SHOW_ARGUMENTS = True


rc.USE_RICH_MARKUP = True
rc.SHOW_ARGUMENTS = True
rc.COLOR_SYSTEM = "truecolor"
rc.STYLE_OPTION = "bold magenta"
rc.STYLE_COMMAND = "bold green"
rc.STYLE_METAVAR = "italic yellow"
rc.STYLE_SWITCH = "underline cyan"
rc.STYLE_USAGE = "bold blue"
rc.STYLE_OPTION_DEFAULT = "dim italic"


def apply_opts(*opts):
  def _wrap(f):
    for opt in reversed(opts):
      f = opt(f)
    return f

  return _wrap


@click.group()
def cli():
  pass


show_figlet()

opt_model = click.option("--model", "-m", required=True, help="Ersilia model id (eosxxxx)")
opt_version = click.option("--version", "-v", default="v1", show_default=True, help="Model version")
opt_project = click.option(
  "--project-name", "-pn", required=False, default=None, help="Project (bucket) name"
)
opt_project_req = click.option("--project-name", "-pn", required=True, help="Project (bucket) name")
opt_input_file = click.option("--input-file", "-i", required=True, help="Path to input CSV")
opt_ins_input_file = click.option("--input-file", "-i", required=False, help="Path to input CSV")
opt_output_file = click.option(
  "--output-file",
  "-o",
  required=False,
  default=None,
  help="Path to output file (csv/h5)",
)
opt_access = click.option(
  "--access",
  type=click.Choice(["both", "public", "private"]),
  default=None,
  required=True,
  show_default=True,
  help="Which buckets to search when project-name not provided",
)
opt_yes_flag = click.option("--yes", "-y", is_flag=True, help="Confirm deletion")
opt_dump_outdir = click.option("--output-dir", "-o", required=False, help="Local output directory")
opt_approx = click.option(
  "--approximate",
  "-nn",
  is_flag=True,
  default=False,
  help="Specifies whether to use Approximate Nearest Neighbor search for result retrieval or not.",
)
opt_cloud = click.option(
  "--cloud",
  "-c",
  is_flag=True,
  default=False,
  help="Specifies to use isaura in cloud mode or not.",
)
opt_start = click.option(
  "--start",
  "-s",
  is_flag=True,
  default=False,
  help="Specifies to start isuara main engines such as minio, milvus, nns server.",
)

opt_isaura_dir = click.option(
  "--isaura-dir",
  "-d",
  required=False,
  default=".",
  show_default=True,
  help="Path to an isaura folder (used mainly to resolve output defaults).",
)

opt_stats_outdir = click.option(
  "--output-dir",
  "-o",
  required=True,
  help="Folder where the stats JSON will be written.",
)


@cli.command("write")
@apply_opts(opt_input_file, opt_project, opt_access, opt_model, opt_version)
def write(input_file, project_name, access, model, version):
  if project_name in (DEFAULT_PRIVATE_BUCKET_NAME, DEFAULT_BUCKET_NAME):
    logger.error("Please specify a different project name. Access denied to write to default project names!")
    sys.exit(1)
  if project_name is None:
    logger.error("Please specify the project name in order to write the data!")
    sys.exit(1)

  with IsauraWriter(
    input_csv=input_file,
    model_id=model,
    model_version=version,
    bucket=project_name,
    access=access,
  ) as w:
    w.write()


@cli.command("read")
@apply_opts(opt_input_file, opt_project, opt_model, opt_version, opt_output_file, opt_approx)
def read(input_file, project_name, model, version, output_file, approximate):
  r = IsauraReader(
    model_id=model, model_version=version, bucket=project_name, input_csv=input_file, approximate=approximate
  )
  r.read(output_csv=output_file)


@cli.command("pull")
@apply_opts(opt_input_file, opt_project, opt_model, opt_version)
def pull(input_file, project_name, model, version):
  pn = project_name or DEFAULT_BUCKET_NAME
  pl = IsauraPull(model_id=model, model_version=version, bucket=pn, input_csv=input_file)
  pl.pull()


@cli.command("push")
@apply_opts(opt_project, opt_model, opt_version)
def push(project_name, model, version):
  p = IsauraPush(model, version, project_name)
  p.push()


@cli.command("copy")
@apply_opts(opt_model, opt_version, opt_project_req, opt_dump_outdir)
def cp(model, version, project_name, output_dir):
  c = IsauraCopy(model_id=model, model_version=version, bucket=project_name, output_dir=output_dir)
  if output_dir is None:
    priv, pub = c.copy()
    logger.info(f"Copied private_new={priv} public_new={pub} from {project_name}")
  else:
    c.copy()


@cli.command("move")
@apply_opts(opt_model, opt_version, opt_project_req)
def mv(model, version, project_name):
  m = IsauraMover(model_id=model, model_version=version, bucket=project_name)
  m.move()
  logger.info(f"Move done for {model}/{version} from {project_name}")


@cli.command("engine")
@apply_opts(opt_start)
def engine(start):
  s = spinner("Starting the engines. Please wait!", run_docker_compose, start)


@cli.command("remove")
@apply_opts(opt_model, opt_version, opt_project_req, opt_yes_flag)
def rm(model, version, project_name, yes):
  if not yes:
    logger.info("Add --yes to confirm deletion")
    sys.exit(1)
  r = IsauraRemover(model_id=model, model_version=version, bucket=project_name)
  r.remove()
  logger.info(f"Remove done for {model}/{version} in {project_name}")


@cli.command("inspect")
@apply_opts(opt_model, opt_version, opt_project, opt_access, opt_ins_input_file, opt_output_file, opt_cloud)
@click.argument("what", type=click.Choice(["inputs"]), required=False, default="inputs")
def cmd_inspect(what, model, version, project_name, access, input_file, output_file, cloud):
  insp = IsauraInspect(
    model_id=model, model_version=version, project_name=project_name, access=access, cloud=cloud
  )
  if input_file:
    df = insp.inspect_inputs(input_file, output_file)
  else:
    df = insp.list_available(output_file)
  logger.info(f"wrote {len(df)} rows{(' -> ' + output_file) if output_file else ''}")


@cli.command("catalog")
@apply_opts(opt_project, opt_cloud)
def cmd_inspect_models(project_name, cloud):
  insp = IsauraInspect(cloud=cloud)
  rows = insp.inspect_models(project_name, prefix_filter="")
  if not rows:
    console.print(f"[yellow]No models found in {project_name}[/]")
    return
  table = make_table(
    f"Model catalog in {project_name}",
    inspect_table,
    rows,
  )
  console.print(table)


@cli.command("stats")
@apply_opts(opt_project, opt_access, opt_cloud, opt_stats_outdir, opt_isaura_dir)
def cmd_stats(project_name, access, cloud, output_dir, isaura_dir):
  ts = datetime.datetime.now(datetime.timezone.utc).strftime("%Y%m%dT%H%M%SZ")
  out_path = os.path.join(output_dir, f"isaura_stats_{ts}.json")

  st = IsauraStat(
    project_name=project_name,
    access=access,
    cloud=cloud,
    endpoint=None,
  )
  written = st.write_json(out_path)
  logger.info(f"isaura stats wrote: {written}")


if __name__ == "__main__":
  cli()



import os, pandas as pd, numpy as np, json, math, os, psutil, requests, subprocess, sys, tempfile, time, yaml
from io import StringIO
from collections import defaultdict
from loguru import logger
from typing import TypeVar
from rich.progress import (
  Progress,
  SpinnerColumn,
  TextColumn,
  BarColumn,
  TimeRemainingColumn,
  TimeElapsedColumn,
)
from rich.text import Text
from rich.table import Table
from rich.logging import RichHandler
from rich.progress import Progress
from rich.console import Console
from rdkit import Chem
from rdkit.Chem import Descriptors, Crippen
from pathlib import Path

try:
  from dotenv import load_dotenv

  load_dotenv(override=False)
except Exception:
  pass

logger.remove()
console = Console()
logger.level("DEBUG", color="<cyan><bold>")
logger.level("INFO", color="<blue><bold>")
logger.level("WARNING", color="<white><bold><bg yellow>")
logger.level("ERROR", color="<white><bold><bg red>")
logger.level("CRITICAL", color="<white><bold><bg red>")
logger.level("SUCCESS", color="<black><bold><bg green>")


ACCESS_FILE = "access.json"
INDEX_FILE = "index.json"
MIN_NNS_RESULT_SIZE = 1_000
MAX_ROWS = 2_000_000

BUILD_STATUS_TIMEOUT = 5
BUILD_START_TIMEOUT = 30
BUILD_POLL_INTERVAL = 0.5
BUILD_MAX_WAIT = 1.0

MW_BINS = [200, 500]
LOGP_BINS = [-1, 5]
GITHUB_ORG = "ersilia-os"
GITHUB_CONTENT_URL = f"https://raw.githubusercontent.com/{GITHUB_ORG}"
METADATA_JSON = "metadata.json"
METADATA_YML = "metadata.yml"


COLLECTION = os.getenv("COLLECTION", "eos3b5e")
MINIO_ENDPOINT = os.getenv("MINIO_ENDPOINT", "http://127.0.0.1:9000")
TIMEOUT = os.getenv("TIMEOUT", 3600)
MINIO_ENDPOINT_CLOUD = os.getenv("MINIO_ENDPOINT_CLOUD") or "http://83.48.73.209:8080"
NNS_ENDPOINT_BASE = os.getenv("NNS_ENDPOINT") or "http://127.0.0.1:8080"
MINIO_LOCAL_AK = os.getenv("MINIO_LOCAL_AK", "minioadmin123")
MINIO_LOCAL_SK = os.getenv("MINIO_LOCAL_SK", "minioadmin1234")
MINIO_CLOUD_AK = os.getenv("MINIO_CLOUD_AK", None)
MINIO_CLOUD_SK = os.getenv("MINIO_CLOUD_SK", None)
MINIO_PRIV_CLOUD_AK = os.getenv("MINIO_PRIV_CLOUD_AK", None)
MINIO_PRIV_CLOUD_SK = os.getenv("MINIO_PRIV_CLOUD_SK", None)

isaura_temp = os.path.join(Path.home(), "isaura", "isaura-temp")
if not os.path.exists(isaura_temp):
  os.makedirs(isaura_temp)

STORE_DIRECTORY = os.getenv("STORE_DIRECTORY", isaura_temp)
MAX_ROWS_PER_FILE = int(os.getenv("MAX_ROWS_PER_FILE", "100000"))
CHECKPOINT_EVERY = int(os.getenv("CHECKPOINT_EVERY", "50000"))
BLOOM_FILENAME = os.getenv("BLOOM_FILENAME", "bloom.pkl")
INPUT_C = ["input", "smiles"]
DEFAULT_BUCKET_NAME = os.getenv("DEFAULT_BUCKET_NAME", "isaura-public")
DEFAULT_PRIVATE_BUCKET_NAME = os.getenv("DEFAULT_PRIVATE_BUCKET_NAME", "isaura-private")
BATCH = int(os.getenv("BATCH", 10_000))
FLUSH_EVERY = os.getenv("FLUSH_EVERY", 10_000)
proc = psutil.Process(os.getpid())

# metdata keys
MKEYS = [
  "Status",
  "Deployment",
  "Source",
  "Source Type",
  "Task",
  "Subtask",
  "Output",
  "Output Dimension",
  "Tag",
  "Biomedical Area",
  "Target Organism",
  "Publication Type",
  "Publication Year",
]

_INT_KEYS = {"Output Dimension", "Publication Year"}

_KEYMAP = {
  "Source Type": "SourceType",
  "Output Dimension": "OutputDimension",
  "Biomedical Area": "BiomedicalArea",
  "Target Organism": "TargetOrganism",
  "Publication Type": "PublicationType",
  "Publication Year": "PublicationYear",
}


# fmt: off
def get(mid, file): return requests.get(f"{GITHUB_CONTENT_URL}/{mid}/main/{file}")
def get_base(mdi, ver): return f"{get_pref(mdi, ver)}/tranches"
def get_desc(pref, wanted): return f"Fetching hive partitions {pref} ({len(wanted)} inputs)"
def get_files_glob(bucket, base): return f"s3://{bucket}/{base}/*/chunk_*.parquet"
def get_keys(file, base): return f"{base}/{file}"
def get_idx_key(base): return get_keys(INDEX_FILE, base)
def get_acc_key(base): return get_keys(ACCESS_FILE,base)
def get_pref(mdi, ver): return f"{mdi}/{ver}"
def get_coll(mdi, ver): return f"{mdi}_{ver}"
def get_params(collection): return {"collection": collection, "batch": str(BATCH), "flush_every": str(FLUSH_EVERY)}
def get_header(): return {"Content-Type": "text/plain"}
def hive_prefix(base): return f"{base}/data"
def make_temp(pref): return tempfile.mkdtemp(prefix=pref, dir=STORE_DIRECTORY)
def rss_mb(): return proc.memory_info().rss / (1024*1024)
def log(msg): logger.info(f"[{time.strftime('%H:%M:%S')}] {msg} | RSS={rss_mb():.1f} MB")
def avail_mem(): return int(psutil.virtual_memory().available)
def mem_gb_lim(ratio=0.8, floor_gb=1): return max(floor_gb, int(avail_mem() * ratio / (1024**3)))
def cpu_cnt(ratio=0.6): return max(1, int(math.floor((os.cpu_count() or 1) * ratio)))
# fmt: on


def pick_meta(d: dict) -> dict:
  def first(v):
    if v is None:
      return None
    if isinstance(v, list):
      return None if not v else v[0]
    return v

  out = {}
  for k in MKEYS:
    v = first(d.get(k))
    kk = _KEYMAP.get(k, k)

    if v is None:
      out[kk] = None
      continue

    if k in _INT_KEYS:
      try:
        out[kk] = int(v)
        continue
      except:
        pass

    out[kk] = str(v)

  return out


def fetch_schema_from_github(model_id):
  try:
    r = get(model_id, METADATA_JSON)
    data = json.load(StringIO(r.text))
  except:
    r = get(model_id, METADATA_YML)
    data = yaml.safe_load(StringIO(r.text))
  return pick_meta(data)


def run_docker_compose(up=True):
  try:
    path = Path(__file__).parent / "configs" / "docker-compose.yml"
    cmd = ["docker", "compose", "-f", path, "up", "-d"] if up else ["docker", "compose", "-f", path, "down"]
    result = subprocess.run(cmd, capture_output=True, text=True, check=True)
    logger.info(f"Docker Compose {'started' if up else 'stopped'} successfully.")
    logger.debug(result.stdout.strip())
    return True
  except subprocess.CalledProcessError as e:
    logger.error(f"Docker Compose failed: {e.stderr.strip()}")
  except FileNotFoundError:
    logger.error("Docker is not installed or not in PATH.")
  except Exception as e:
    logger.error(f"Unexpected error: {e}")
  return False


def show_figlet():
  path = Path(__file__).parent / "assets" / "figlet.txt"
  text = Path(path).read_text(encoding="utf-8")
  start_color = (0, 255, 255)
  end_color = (255, 0, 255)
  content = "".join(text)
  gradient = Text()
  for i, ch in enumerate(content):
    ratio = i / max(1, len(content) - 1)
    r = int(start_color[0] + (end_color[0] - start_color[0]) * ratio)
    g = int(start_color[1] + (end_color[1] - start_color[1]) * ratio)
    b = int(start_color[2] + (end_color[2] - start_color[2]) * ratio)
    gradient.append(ch, style=f"rgb({r},{g},{b})")
  print()
  console.print(gradient, justify="center")
  console.print(Text(f"[🚀] [New] Version 2.1.16", style="bold bright_black"), justify="center")
  print()


def split_csv(df):
  paths = []
  output_dir = make_temp("isaura_push_")
  for bucket in [DEFAULT_BUCKET_NAME, DEFAULT_PRIVATE_BUCKET_NAME]:
    if bucket in df["bucket"].unique():
      path = os.path.join(output_dir, f"{bucket.replace('-', '_')}.csv")
      df[df["bucket"] == bucket].to_csv(path, index=False)
      paths.append(str(path))
  return paths


def query(conn, header, wanted, file_glob, columns="*", tmpdir="/tmp"):
  if not wanted:
    return pd.DataFrame()
  try:
    conn.execute(f"SET memory_limit='{mem_gb_lim()}GB'")
    conn.execute(f"SET temp_directory='{tmpdir}'")
    conn.execute("PRAGMA enable_object_cache")
    conn.execute(f"SET threads TO {cpu_cnt()}")
  except Exception:
    pass
  wanted_list = list(wanted)
  order = np.arange(len(wanted_list), dtype=np.int64)
  wdf = pd.DataFrame({header: wanted_list, "__o": order})
  conn.register("wanted_inputs", wdf)
  sql = f"""
        WITH p AS (
          SELECT {columns}
          FROM read_parquet('{file_glob}')
          WHERE {header} IN (SELECT {header} FROM wanted_inputs)
        )
        SELECT p.*
        FROM p
        JOIN wanted_inputs w
          ON p.{header} = w.{header}
        ORDER BY w.__o
    """
  try:
    out = conn.execute(sql).fetchdf()
  finally:
    conn.unregister("wanted_inputs")
  return out


def group_inputs(wanted, index, bloom=None, force=False):
  logger.debug(f"Checking {len(wanted)} inputs in the {len(index)} index!")
  try:
    if bloom:
      miss = [s for s in wanted if not bloom.seen(s)]
    g = defaultdict(set)
    if miss and not force:
      logger.error(
        f"inputs not indexed: {miss[:5]}{'...' if len(miss) > 5 else ''} total_missing={len(miss)}"
      )
      sys.exit(1)
    if not force:
      for s in wanted:
        r, c = index[s]
        g[(int(r), int(c))].add(s)
      return g
    if miss and force:
      for s in miss:
        r, c, _, _ = tranche_coordinates(s)
        g[(int(r), int(c))].add(s)
      return g
  except Exception as e:
    logger.error(e)
    return None


def line_gen(df, chunksize=100_000):
  for start in range(0, len(df), chunksize):
    chunk = df.iloc[start : start + chunksize]
    for s in chunk["input"].astype(str):
      yield (s + "\n").encode("utf-8")
    log(f"sent chunk rows={len(chunk):,}")
  log("finished streaming")


def build_index_status(collection):
  r = requests.get(
    f"{NNS_ENDPOINT_BASE}/build_index",
    params={"collection": collection},
    timeout=BUILD_STATUS_TIMEOUT,
  )
  r.raise_for_status()
  return r.json()


def start_build_index(collection, nlist=None, rebuild=False, wait=False):
  params = {"collection": collection}
  if nlist is not None:
    params["nlist"] = str(nlist)
  if rebuild:
    params["rebuild"] = "1"
  if wait:
    params["wait"] = "1"
  r = requests.post(
    f"{NNS_ENDPOINT_BASE}/build_index",
    params=params,
    timeout=BUILD_START_TIMEOUT,
  )
  r.raise_for_status()
  return r.json()


# Approximare search helpers
def ensure_index_ready(collection, max_wait_s=BUILD_MAX_WAIT):
  st = build_index_status(collection)

  if not st.get("exists", False):
    try:
      start_build_index(collection, wait=False)
    except requests.RequestException as e:
      return False, {"error": f"start_build_index_failed: {e}", "status": st}

    st = build_index_status(collection)

  if st.get("is_failed", False):
    return False, {"error": "index_failed", "status": st}

  if st.get("is_finished", False) and not st.get("is_building", False):
    return True, {"status": st}

  t_end = time.time() + max_wait_s
  while time.time() < t_end:
    time.sleep(BUILD_POLL_INTERVAL)
    st = build_index_status(collection)
    if st.get("is_failed", False):
      return False, {"error": "index_failed", "status": st}
    if st.get("is_finished", False) and not st.get("is_building", False):
      return True, {"status": st}

  return False, {"error": "index_building", "status": st}


def post_apprx(df, collection, nlist=None):
  t0 = time.time()
  try:
    with requests.Session() as s:
      log("start streaming to nns api. This process sometimes appears to be slow. Please have some patience!")
      resp = s.post(
        f"{NNS_ENDPOINT_BASE}/insert",
        params=get_params(collection),
        data=line_gen(df),
        headers=get_header(),
        timeout=None,
      )
    dt = (time.time() - t0) * 1000
    log(f"done total_ms={dt:.0f}. Body: {resp.text[:1000]}")

    try:
      start_build_index(collection, nlist=nlist, rebuild=False, wait=False)
      st = build_index_status(collection)
      logger.info(
        f"index build triggered collection={collection} exists={st.get('exists')} "
        f"building={st.get('is_building')} finished={st.get('is_finished')} "
        f"progress={st.get('progress_pct')}"
      )
    except requests.RequestException as e:
      logger.error(f"index build trigger failed collection={collection}: {e}")

    return resp
  except requests.RequestException as e:
    logger.error(f"approx NN search failed. The NNS server container may not be sarted!: {e}")
    return None


def get_apprx(inputs, collection, fallback_search=None):
  try:
    logger.info(f"Sending {len(inputs)} inputs for ANN search server to get top 1 similar compounds")

    ok, meta = ensure_index_ready(collection, max_wait_s=BUILD_MAX_WAIT)
    if not ok:
      st = meta.get("status", {})
      err = meta.get("error", "index_not_ready")
      msg = (
        f"ANN index not ready collection={collection} error={err} "
        f"exists={st.get('exists')} building={st.get('is_building')} "
        f"finished={st.get('is_finished')} failed={st.get('is_failed')} "
        f"progress={st.get('progress_pct')}"
      )
      logger.error(msg)
      if callable(fallback_search):
        logger.info("Please use conventional/exact search")
        return fallback_search(inputs, collection)
      return []

    r = requests.post(
      f"{NNS_ENDPOINT_BASE}/search",
      json={"collection": collection, "smiles": inputs},
      timeout=TIMEOUT,
    )
    r.raise_for_status()

    payload = r.json() or {}
    results = payload.get("results", [])
    return [x.get("match") for x in results if "match" in x]

  except requests.RequestException as e:
    logger.error(f"approx NN search failed. The NNS server container may not be sarted!: {e}")
    if callable(fallback_search):
      logger.info("Falling back to conventional/exact search")
      return fallback_search(inputs, collection)
    return []


def write_access_file(existed, data, access, dir):
  try:
    if data:
      m = [{"input": i, "access": access} for i in data]
      if existed:
        m = m + existed
      with open(dir, "w") as f:
        json.dump(m, f, indent=2)
  except Exception as e:
    logger.error(e)


def tranche_coordinates(smiles):
  mol = Chem.MolFromSmiles(smiles)
  if mol is None:
    raise ValueError("Invalid SMILES")
  mw, logp = Descriptors.MolWt(mol), Crippen.MolLogP(mol)

  for i, edge in enumerate(MW_BINS):
    if mw <= edge:
      col = i + 1
      break
  else:
    col = len(MW_BINS) + 1

  for j, edge in enumerate(LOGP_BINS):
    if logp <= edge:
      row = j + 1
      break
  else:
    row = len(LOGP_BINS) + 1

  return row, col, mw, logp


# Table + progress bar helpers
def make_table(title, cols, rows):
  t = Table(title=title)
  for c in cols:
    t.add_column(c["name"], justify=c.get("justify", "left"), style=c.get("style", ""))
  for r in rows:
    t.add_row(*[str(r.get(c["key"], "")) for c in cols])
  return t


inspect_table = [
  {"key": "model", "name": "model/version", "justify": "left", "style": "bold"},
  {"key": "entries", "name": "entry count", "justify": "right"},
  {"key": "chunks", "name": "chunks", "justify": "right"},
]

T = TypeVar("T")


def track_write_progress(rows, total=None, description="Writing rows", console=None):
  if total is None:
    progress = Progress(
      SpinnerColumn(),
      TextColumn("[progress.description]{task.description}"),
      TextColumn("{task.completed} rows"),
      TimeElapsedColumn(),
      console=console,
      transient=True,
    )
  else:
    progress = Progress(
      SpinnerColumn(),
      TextColumn("[progress.description]{task.description}"),
      BarColumn(),
      TextColumn("{task.completed}/{task.total}"),
      TimeElapsedColumn(),
      TimeRemainingColumn(),
      console=console,
      transient=True,
    )

  with progress:
    task_id = progress.add_task(description, total=total)
    for row in rows:
      yield row
      progress.advance(task_id, 1)


def spinner(message, fn, *args, **kwargs):
  c = Console()
  with c.status(message, spinner="dots"):
    result = fn(*args, **kwargs)
  return result


# fmt: off
class Logger:
  def __init__(self):
    self.logger = logger
    self._console = None
    self._file = None
    self._log_to_console()

  def _log_to_console(self):
    if self._console is None:
      rich_handler = RichHandler(
        rich_tracebacks=True, markup=True, log_time_format="%H:%M:%S", show_path=False
      )
      self._rich_console = rich_handler.console

      self._console = self.logger.add(
        rich_handler,
        format="{message}",
        colorize=True,
      )

  @property
  def console(self):
    return self._rich_console
  
  def _unlog_from_console(self):
    if self._console is not None:
      try:
        self.logger.remove(self._console)
      except Exception:
        pass
      self._console = None

  def set_verbosity(self, verbose):
    if verbose:
      self._log_to_console()
    else:
      self._unlog_from_console()

  def debug(self, text): self.logger.debug(text)
  def info(self, text): self.logger.info(text)
  def warning(self, text): self.logger.warning(text)
  def error(self, text): self.logger.error(text)
  def critical(self, text): self.logger.critical(text)
  def success(self, text): self.logger.success(text)


logger = Logger()
